<?php

namespace Drupal\feeds_filemaker\Feeds\Fetcher;

use Drupal\Core\State\StateInterface;
use Drupal\feeds\Exception\EmptyFeedException;
use Drupal\feeds\Feeds\Fetcher\HttpFetcher;
use Drupal\feeds\Result\HttpFetcherResult;
use Drupal\feeds\FeedInterface;
use Drupal\feeds\StateInterface as FeedsStateInterface;
use GuzzleHttp\Exception\RequestException;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\key\KeyRepositoryInterface;

/**
 * @FeedsFetcher(
 *   id = "filemaker_api_fetcher",
 *   title = @Translation("FileMaker API Fetcher"),
 *   description = @Translation("Fetches data from FileMaker API with authentication."),
 *   form = {
 *     "configuration" = "Drupal\feeds_filemaker\Feeds\Fetcher\Form\FileMakerApiFetcherForm",
 *     "feed" = "Drupal\feeds\Feeds\Fetcher\Form\HttpFetcherFeedForm",
 *   }
 * )
 */
class FileMakerApiFetcher extends HttpFetcher implements ContainerFactoryPluginInterface {

  protected KeyRepositoryInterface $keyRepository;
  protected StateInterface $coreState;

  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->setKeyRepository($container->get('key.repository'));
    $instance->setCoreState($container->get('state'));
    return $instance;
  }

  public function setKeyRepository(KeyRepositoryInterface $key_repository): self {
    $this->keyRepository = $key_repository;
    return $this;
  }

  public function setCoreState(StateInterface $state): self {
    $this->coreState = $state;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  /**
   * {@inheritdoc}
   */
  /**
   * {@inheritdoc}
   */
  public function fetch(FeedInterface $feed, FeedsStateInterface $state) {
    $sink = $this->feedsFileSystem->tempnam($feed, 'filemaker_fetcher_');
    $source = $feed->getSource();

    $offset_key = 'filemaker_' . $feed->id() . '_offset';
    $offset = $this->coreState->get($offset_key, 1);  // Start from 1 instead of 0
    $limit = (int) $this->configuration['batch_size'];
    $total_items = 0;
    $all_items = [];

    try {
      $token = $this->authenticate();

      do {
        $options = [
          'headers' => [
            'Authorization' => 'Bearer ' . $token,
          ],
          'query' => [
            '_limit' => $limit,
            '_offset' => $offset,
          ],
        ];

        $response = $this->client->get($source, $options);
        $body = json_decode($response->getBody(), TRUE);

        // Parse the items from the correct location in the response
        $items = $body['response']['data'] ?? [];
        $item_count = count($items);

        if ($item_count === 0) {
          break;
        }

        // Merge the new items with the existing items
        $all_items = array_merge($all_items, $items);

        $total_items += $item_count;
        $offset += $item_count;  // Increment by actual number of items received

        // Update the state with the new offset
        $this->coreState->set($offset_key, $offset);

      } while ($item_count === $limit);

      if ($total_items === 0) {
        $this->fileSystem->unlink($sink);
        $state->setMessage($this->t('The feed is empty.'));
        throw new EmptyFeedException();
      }

      // Write all items to the sink file as a single JSON array
      file_put_contents($sink, json_encode($all_items));

      // Reset the offset for the next import
      $this->coreState->set($offset_key, 1);  // Reset to 1 instead of 0

      return new HttpFetcherResult($sink, $response->getHeaders(), $this->fileSystem);
    }
    catch (RequestException $e) {
      $this->fileSystem->unlink($sink);
      $state->setMessage($this->t('The FileMaker API request failed with error: @error', [
        '@error' => $e->getMessage(),
      ]));
      throw new EmptyFeedException();
    }
  }

  protected function authenticate() {
    $auth_endpoint = $this->getCredential('filemaker_auth_endpoint');
    $username = $this->getCredential('filemaker_username');
    $password = $this->getCredential('filemaker_password');

    $response = $this->client->post($auth_endpoint, [
      'headers' => [
        'Authorization' => 'Basic ' . base64_encode("$username:$password"),
      ],
      'json' => [
        'fmDataSource' => [],
      ],
    ]);

    $body = json_decode($response->getBody(), TRUE);
    return $body['response']['token'];
  }

  protected function getCredential(string $key_id): string {
    $key = $this->keyRepository->getKey($key_id);
    if (!$key) {
      throw new \RuntimeException("Key '$key_id' not found.");
    }
    return $key->getKeyValue();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
        'batch_size' => 100,
      ] + parent::defaultConfiguration();
  }
}
